// Etcd client mock for testing purposes. It tries to follow the same
// semantics as the actual etcd, at least as far as they are used by
// autoradio. Instead of talking to a remote etcd server, it uses an
// in-memory representation of the data.

package util

import (
	"errors"
	"math/rand"
	"strings"
	"sync"
	"time"

	"git.autistici.org/ale/autoradio"
	"github.com/coreos/go-etcd/etcd"
)

type datum struct {
	value  string
	expire time.Time
}

type testEtcdServer struct {
	lock    sync.Mutex
	latency time.Duration
	data    map[string]datum
	watches map[string][]chan *etcd.Response
	index   uint64
}

func NewTestEtcdClient() autoradio.EtcdClient {
	return &testEtcdServer{
		data:    make(map[string]datum),
		watches: make(map[string][]chan *etcd.Response),
		index:   1,
	}
}

func NewTestEtcdClientWithLatency(maxLatency time.Duration) autoradio.EtcdClient {
	return &testEtcdServer{
		data:    make(map[string]datum),
		watches: make(map[string][]chan *etcd.Response),
		index:   1,
		latency: maxLatency,
	}
}

func (s *testEtcdServer) delay() {
	if s.latency > 0 {
		time.Sleep(time.Duration(rand.Int63n(int64(s.latency))))
	}
}

func (s *testEtcdServer) trigger(action, key string) *etcd.Response {
	resp := &etcd.Response{
		Action:    action,
		EtcdIndex: s.index,
		Node: &etcd.Node{
			Key: key,
		},
	}
	if action != "delete" {
		resp.Node.Value = s.data[key].value
	}
	for pfx, w := range s.watches {
		if strings.HasPrefix(key, pfx) {
			for _, ch := range w {
				ch <- resp
			}
		}
	}
	s.index++
	return resp
}

func (s *testEtcdServer) Create(key, value string, ttl uint64) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	defer s.lock.Unlock()

	if _, ok := s.data[key]; ok {
		return nil, &etcd.EtcdError{Message: "already there", Index: s.index}
	}

	s.data[key] = datum{
		value:  value,
		expire: time.Now().Add(time.Duration(ttl) * time.Second),
	}
	return s.trigger("create", key), nil
}

func (s *testEtcdServer) CompareAndSwap(key, value string, ttl uint64, oldvalue string, index uint64) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	defer s.lock.Unlock()
	if d, ok := s.data[key]; ok && d.value == oldvalue {
		s.data[key] = datum{
			value:  value,
			expire: time.Now().Add(time.Duration(ttl) * time.Second),
		}
		return s.trigger("update", key), nil
	}
	return nil, &etcd.EtcdError{Message: "failed", Index: s.index}
}

func (s *testEtcdServer) Delete(key string, recursive bool) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	defer s.lock.Unlock()
	delete(s.data, key)
	return s.trigger("delete", key), nil
}

func (s *testEtcdServer) Get(key string, recursive, boh bool) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	defer s.lock.Unlock()
	resp := &etcd.Response{
		EtcdIndex: s.index,
	}
	var nodes []*etcd.Node
	keyDirPfx := key + "/"
	for path, datum := range s.data {
		if path == key || strings.HasPrefix(path, keyDirPfx) {
			nodes = append(nodes, &etcd.Node{
				Key:   path,
				Value: datum.value,
			})
		}
	}
	switch {
	case len(nodes) == 0:
		return nil, errors.New("not found")
	case len(nodes) == 1 && nodes[0].Key == key:
		resp.Node = nodes[0]
	default:
		resp.Node = &etcd.Node{
			Key:   key,
			Dir:   true,
			Nodes: nodes,
		}
	}
	return resp, nil
}

func (s *testEtcdServer) Set(key, value string, ttl uint64) (*etcd.Response, error) {
	s.delay()
	s.lock.Lock()
	defer s.lock.Unlock()
	s.data[key] = datum{
		value:  value,
		expire: time.Now().Add(time.Duration(ttl) * time.Second),
	}
	return s.trigger("set", key), nil
}

func (s *testEtcdServer) Watch(key string, index uint64, recursive bool, respch chan *etcd.Response, stop chan bool) (*etcd.Response, error) {
	ch := respch
	if ch == nil {
		ch = make(chan *etcd.Response, 1)
	}
	s.lock.Lock()
	s.watches[key] = append(s.watches[key], ch)
	s.lock.Unlock()

	var resp *etcd.Response
	if respch != nil {
		<-stop
	} else {
		select {
		case resp = <-ch:
		case <-stop:
		}
	}

	// Delete the watch.
	s.lock.Lock()
	var watches []chan *etcd.Response
	for _, w := range s.watches[key] {
		if w != ch {
			watches = append(watches, w)
		}
	}
	s.watches[key] = watches
	s.lock.Unlock()
	close(ch)

	if resp == nil {
		return nil, etcd.ErrWatchStoppedByUser
	}
	return resp, nil
}
